package io.redvox.api900;

import java.util.*;

/**
 * This class contains utility functions for data conversion and statistics.
 */
public class Util {
    /**
     * Converts a list with an even number of elements into a key-pair map by changing every pair of values in the list
     * into a key-pair within the map.
     *
     * As an example, the list [1, 2, 3, 4, 5, 6] would be converted into the map {1:2, 3:4, 5:6}
     *
     * @param list A list containing an even number of values.
     * @param <T> The generic type of elements in the list (can be any type).
     * @return A map containing key-pair values generated by taking adjacent pairs of values from a list.
     */
    public static <T> Map<T, T> listToKeyPair(final List<T> list) {
        if(list.size() % 2 == 1) {
            throw Api900Exception.of("List must contain even number of items");
        }

        final Map<T, T> keyPairs = new HashMap<>();
        final ListIterator<T> iterator = list.listIterator();
        while(iterator.hasNext()) {
            keyPairs.put(iterator.next(), iterator.next());
        }
        return keyPairs;
    }

    /**
     * Returns the mean value of a list of numbers.
     * @param numbers A list of values that extends Number.
     * @param <T> A type that extends Number.
     * @return The mean of a list of numbers.
     */
    public static <T extends Number> double mean(final List<T> numbers) {
        if(numbers.size() == 0) {
            throw Api900Exception.of("mean requires at least 1 sample point");
        }
        double sum = 0;
        for(final T v: numbers) {
            sum += v.doubleValue();
        }
        return sum / numbers.size();
    }

    /**
     * Returns the median value of a list of numbers.
     * @param numbers A list of values that extends Number.
     * @param <T> A type that extends Number.
     * @return The median of a list of numbers.
     */
    public static <T extends Number & Comparable<T>> double median(final List<T> numbers) {
        if(numbers.size() == 0) {
            throw Api900Exception.of("median requires at least 1 sample point");
        }
        Collections.sort(numbers);
        final int middle = numbers.size() / 2;
        if (numbers.size() % 2 == 1) {
            return numbers.get(middle).doubleValue();
        } else {
            return (numbers.get(middle-1).doubleValue() + numbers.get(middle).doubleValue()) / 2.0;
        }
    }

    /**
     * Returns the variance value of a list of numbers.
     * @param numbers A list of values that extends Number.
     * @param <T> A type that extends Number.
     * @return The variance of a list of numbers.
     */
    public static <T extends Number> double variance(final List<T> numbers) {
        if(numbers.size() < 2) {
            throw Api900Exception.of("variance requires at least 2 sample points");
        }
        final double mean = mean(numbers);
        double sqSum = 0;
        double d;
        for(final T v: numbers) {
            d = v.doubleValue();
            sqSum += (d - mean) * (d - mean);
        }
        return sqSum / numbers.size();
    }

    /**
     * Returns the stddev value of a list of numbers.
     * @param numbers A list of values that extends Number.
     * @param <T> A type that extends Number.
     * @return The stddev of a list of numbers.
     */
    public static <T extends Number> double stddev(final List<T> numbers) {
        if(numbers.size() < 2) {
            throw Api900Exception.of("stddev requires at least 2 sample points");
        }
        return Math.sqrt(variance(numbers));
    }
}
